def _gdecl(*parts):
    not_none = (part for part in parts if part is not None)
    return '{};'.format(' '.join(not_none))


def location_str(location):
    if location is None:
        return None
    else:
        return 'layout(location={})'.format(int(location))


class GlslVar(object):
    """Represent a GLSL variable declaration (or struct member)."""
    def __init__(self, name, gtype, interpolation=None):
        self.name = name
        self.gtype = gtype
        self.interpolation = interpolation

    def declare(self):
        return _gdecl(self.interpolation, self.gtype, self.name)

    def declare_uniform(self):
        return _gdecl('uniform', self.gtype, self.name)

    def declare_attribute(self, location=None):
        return _gdecl(location_str(location), self.interpolation,
                      'in', self.gtype, self.name)

    def declare_output(self, location=None):
        return _gdecl(self.interpolation, location_str(location),
                      'out', self.gtype, self.name)



from unittest import TestCase

from shaderdef.glsl_var import GlslVar

class TestGlslVar(TestCase):
    def test_declare(self):
        var = GlslVar('x', 'vec3')
        self.assertEqual(var.declare(), 'vec3 x;')

    def test_declare_uniform(self):
        var = GlslVar('x', 'vec3')
        self.assertEqual(var.declare_uniform(), 'uniform vec3 x;')

    def test_declare_attribute(self):
        var = GlslVar('x', 'vec3')
        self.assertEqual(var.declare_attribute(), 'in vec3 x;')

    def test_declare_output(self):
        var = GlslVar('x', 'vec3')
        self.assertEqual(var.declare_output(), 'out vec3 x;')

    def test_declare_interpolation(self):
        var = GlslVar('x', 'vec3', interpolation='noperspective')
        self.assertEqual(var.declare_attribute(), 'noperspective in vec3 x;')



from typing import Iterator, Sequence, get_type_hints

from shaderdef.ast_util import (parse_source,
                                remove_function_parameters,
                                remove_function_return_type,
                                rename_function)
from shaderdef.find_function import find_function
from shaderdef.interface import AttributeBlock, FragmentShaderOutputBlock
from shaderdef.lift_attributes import lift_attributes
from shaderdef.rename_ast_nodes import rename_gl_builtins
from shaderdef.rewrite_output import rewrite_return_as_assignments
from shaderdef.py_to_glsl import py_to_glsl


def make_prefix(name):
    parts = name.split('_')
    return ''.join(part[0] for part in parts) + '_'


def get_output_interface(func):
    return_type = get_type_hints(func).get('return')
    if return_type is None:
        return None
    # Unwrap iterators (used for geom shader output)
    origin = getattr(return_type, '__origin__', None)
    if origin is not None and origin == Iterator:
        return_type = return_type.__parameters__[0]
    return return_type


class Stage(object):
    def __init__(self, func):
        self.name = func.__name__
        root = parse_source(func)
        self.ast_root = find_function(root, self.name)
        self.input_prefix = ''
        self.output_prefix = make_prefix(self.name)
        self._glsl_source = None
        # TODO
        self._params = get_type_hints(func)
        self._params.pop('return', None)
        self._return_type = get_output_interface(func)

    def declare_inputs(self, lines):
        for name, param_type in self._params.items():
            # TODO(nicholasbishop): dedup with return type
            origin = getattr(param_type, '__origin__', None)
            array = None
            if origin is not None and origin == Sequence:
                param_type = param_type.__parameters__[0]
                array = True
            lines += param_type.declare_input_block(instance_name=name,
                                                    array=array)

    @staticmethod
    def define_aux_functions(lines, library):
        # TODO(nicholasbishop): for now we don't attempt to check if
        # the function is actually used, just define them all
        for func in library:
            func_node = parse_source(func)
            lines += py_to_glsl(func_node)

    @property
    def glsl(self):
        if self._glsl_source is None:
            raise ValueError('shader has not been translated yet')
        return self._glsl_source

    def translate(self, library):
        self._glsl_source = self.to_glsl(library)

    def apply_decorators(self):
        decs = self.ast_root.decorator_list
        if len(decs) == 1 and decs[0].func.id == 'geom_shader_meta':
            kwargs = {}
            for keyword in decs[0].keywords:
                kwargs[keyword.arg] = keyword.value

            # TODO(nicholasbishop): validate inputs
            input_primitive = kwargs['input_primitive'].id
            output_primitive = kwargs['output_primitive'].id
            max_vertices = int(kwargs['max_vertices'].n)

            yield 'layout({}) in;'.format(input_primitive)
            yield 'layout({}, max_vertices = {}) out;'.format(output_primitive,
                                                              max_vertices)

    def attributes_to_lift(self):
        # Attributes aren't allowed in an interface block
        for param_name, param_type in self._params.items():
            if issubclass(param_type, AttributeBlock):
                yield param_name

        # Same for fragment shader outputs. TODO(nicholasbishop): we
        # could also do this more directly during rewrite_return...
        if self._return_type is not None:
            if issubclass(self._return_type, FragmentShaderOutputBlock):
                yield self._return_type.instance_name()

    def to_glsl(self, library):
        lines = []
        lines.append('#version 330 core')

        lines += list(self.apply_decorators())
        self.declare_inputs(lines)
        self.define_aux_functions(lines, library)

        ast_root = self.ast_root

        # The main shader function must always be "void main()"
        rename_function(ast_root, 'main')
        remove_function_parameters(ast_root)
        remove_function_return_type(ast_root)

        ast_root = rewrite_return_as_assignments(ast_root, self._return_type)
        ast_root = lift_attributes(ast_root, set(self.attributes_to_lift()))
        ast_root = rename_gl_builtins(ast_root)

        if self._return_type is not None:
            lines += self._return_type.declare_output_block()

        lines += py_to_glsl(ast_root)
        return '\n'.join(lines)



# pylint: disable=missing-docstring

import ast
from unittest import TestCase

from shaderdef.ast_util import parse_source, remove_function_parameters
from shaderdef.attr_rename import rename_attributes
from shaderdef.find_function import find_function
from test.util import AstTestCase


class SimpleClass(object):
    """Do-nothing class used in some of the tests."""
    def my_method(self):
        pass

    def method_with_params(self, thing: int):
        return self, thing


class TestFindFunction(TestCase):
    def setUp(self):
        self.root = parse_source(SimpleClass)

    def test_find_function(self):
        self.assertIsNot(find_function(self.root, 'my_method'), None)

    def test_function_not_found(self):
        with self.assertRaises(KeyError):
            find_function(self.root, 'bad_name')


class TestAttrRename(AstTestCase):
    def test_load(self):
        root = rename_attributes(ast.parse('var = self.value1'),
                                 load_names={'value1': 'value2'})
        expected = ast.parse('var = self.value2')
        self.assertEqual(root, expected)

    def test_store(self):
        root = rename_attributes(ast.parse('self.var1 = value'),
                                 store_names={'var1': 'var2'})
        expected = ast.parse('self.var2 = value')
        self.assertEqual(root, expected)

    def test_load_and_store(self):
        root = rename_attributes(ast.parse('self.var1 = self.value1'),
                                 load_names={'value1': 'value2'},
                                 store_names={'var1': 'var2'})
        expected = ast.parse('self.var2 = self.value2')
        self.assertEqual(root, expected)

    def test_load_and_store_same_key(self):
        root = rename_attributes(ast.parse('self.foo1 = self.foo1'),
                                 load_names={'foo1': 'load'},
                                 store_names={'foo1': 'store'})
        expected = ast.parse('self.store = self.load')
        self.assertEqual(root, expected)

    def test_ignore_function_call(self):
        root = rename_attributes(ast.parse('self.myfunc(1)'),
                                 load_names={'myfunc': 'load'},
                                 store_names={'myfunc': 'store'})
        expected = ast.parse('self.myfunc(1)')
        self.assertEqual(root, expected)


class TestRemoveFunctionParameters(AstTestCase):
    def setUp(self):
        self.root = parse_source(SimpleClass)

    def test_remove_all_params(self):
        actual = find_function(self.root, 'method_with_params')
        remove_function_parameters(actual)

        root = ast.parse('def method_with_params(): return self, thing')
        expected = find_function(root, 'method_with_params')

        self.assertEqual(actual, expected)



from shaderdef.glsl_var import GlslVar
from shaderdef.interface import (AttributeBlock,
                                 FragmentShaderOutputBlock, GlGsIn,
                                 ShaderInterface, UniformBlock)
from shaderdef.shader import ShaderDef



from unittest import TestCase

from shaderdef.glsl_types import vec4
from shaderdef.interface import ShaderInterface
from shaderdef.shader import ShaderDef
from test.util import deindent


class MyMaterial(object):
    # pylint: disable=no-self-use

    class FsOut(ShaderInterface):
        color = vec4()

    def vert_shader(self):
        pass

    def geom_shader(self):
        pass

    def frag_shader(self) -> 'MyMaterial.FsOut':
        return MyMaterial.FsOut(color=vec4(1.0, 0.0, 0.0, 1.0))


class TestShader(TestCase):
    def test_simple(self):
        shader = ShaderDef(MyMaterial.vert_shader,
                           MyMaterial.geom_shader,
                           MyMaterial.frag_shader)
        shader.translate()
        self.assertEqual(deindent(shader.vert_shader),
                         '#version 330 core'
                         'void main() {}')
        self.assertEqual(deindent(shader.frag_shader),
                         '#version 330 core'
                         'out FsOut {'
                         'vec4 color;'
                         '} fs_out;'
                         'void main() {fs_out.color = vec4(1.0, 0.0, 0.0, 1.0);}')



# pylint: disable=unused-argument

def geom_shader_meta(input_primitive, output_primitive, max_vertices):
    def passthrough(func):
        return func
    return passthrough

def end_primitive():
    pass

def exp2(var):
    pass

def mod(num1, num2):
    pass

def length(vec_type):
    pass



"""Version utilities that don't rely on importing shaderdef."""

import os

SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
VERSION_PATH = os.path.join(SCRIPT_DIR, '../shaderdef/version.py')

def parse_version_string(version_string):
    """Parse dotted version string into a list of ints."""
    parts = version_string.split('.')
    return [int(part) for part in parts]


def format_version_string(version):
    """Format a list of version components."""
    return '.'.join(str(part) for part in version)


def load_version_as_string():
    """Get the current version from version.py as a string."""
    with open(VERSION_PATH, 'r') as rfile:
        contents = rfile.read().strip()

    _, version = contents.split('=')
    version = version.strip()

    # Remove quotes
    return version.strip('"\'')


def load_version_as_list():
    """Get the current version from version.py as a list of ints."""
    return parse_version_string(load_version_as_string())



import ast

def kwargs_as_assignments(call_node, parent):
    """Yield Assign nodes from kwargs in a Call node."""
    if not isinstance(call_node, ast.Call):
        raise TypeError('node must be an ast.Call')

    if len(call_node.args) > 0:
        raise ValueError('positional args not allowed')

    for keyword in call_node.keywords:
        dst_name = keyword.arg

        if dst_name.startswith('gl_'):
            # Write to builtins directly
            target = [ast.Name(id=keyword.arg, ctx=ast.Load())]
        else:
            # Non-builtins are part of an interface block
            target = [ast.Attribute(value=parent, attr=keyword.arg,
                                    ctx=ast.Store())]

        yield ast.Assign(targets=target, value=keyword.value)


class _RewriteReturn(ast.NodeTransformer):
    def __init__(self, interface):
        self.interface = interface

    def _output_to_list(self, node):
        parent = ast.Name(id=self.interface.instance_name(), ctx=ast.Load())
        return list(kwargs_as_assignments(node.value, parent))

    def visit_Return(self, node):  # pylint: disable=invalid-name
        return self._output_to_list(node)

    def visit_Expr(self, node):  # pylint: disable=invalid-name
        if isinstance(node.value, ast.Yield):
            lst = self._output_to_list(node.value)
            lst.append(ast.parse('EmitVertex()'))
            return lst
        else:
            return node


def rewrite_return_as_assignments(func_node, interface):
    """Modify FunctionDef node to directly assign instead of return."""
    func_node = _RewriteReturn(interface).visit(func_node)
    ast.fix_missing_locations(func_node)
    return func_node



import ast

class AttrRename(ast.NodeTransformer):
    # pylint: disable=invalid-name
    def __init__(self, load_names, store_names, call_names):
        self._load_names = load_names
        self._store_names = store_names
        self._call_names = call_names
        self._func_attrs = set()

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute):
            self._func_attrs.add(node.func)
        return self.generic_visit(node)

    def visit_Attribute(self, node):
        # Don't transform function names
        if node in self._func_attrs:
            names = self._call_names
        elif isinstance(node.ctx, ast.Load):
            names = self._load_names
        elif isinstance(node.ctx, ast.Store):
            names = self._store_names

        new_name = names.get(node.attr)
        if new_name is not None:
            node.attr = new_name

        return node


def rename_attributes(root, load_names=None, store_names=None,
                      call_names=None):
    if load_names is None:
        load_names = {}
    if store_names is None:
        store_names = {}
    if call_names is None:
        call_names = {}
    return AttrRename(load_names, store_names, call_names).visit(root)



import ast
from unittest import TestCase

from shaderdef.py_to_glsl import py_to_glsl
from test.util import deindent

class TestPyToGlsl(TestCase):
    def assert_code_equal(self, pycode, glslcode):
        root = ast.parse(pycode)
        code = '\n'.join(py_to_glsl(root))
        self.assertEqual(deindent(code), glslcode)

    def test_empty_function(self):
        self.assert_code_equal('def myFunc(): pass', 'void myFunc() {}')

    def test_assign(self):
        self.assert_code_equal('a = b', 'a = b')

    def test_binop(self):
        self.assert_code_equal('a - b * c / d', '(a - ((b * c) / d))')

    def test_subscript(self):
        self.assert_code_equal('a[0]', 'a[0]')

    def test_unaryop(self):
        self.assert_code_equal('-a', '-a')

    def test_augassign(self):
        self.assert_code_equal('a += 1', 'a += 1')

    def test_compare(self):
        self.assert_code_equal('a < b', 'a < b')

    def test_if(self):
        self.assert_code_equal('if a: a = b', 'if (a) {a = b;}')

    def test_if_else(self):
        self.assert_code_equal('if a:\n return 1\n'
                               'else:\n return 2\n',
                               'if (a) {return 1;} else {return 2;}')

    def test_for(self):
        self.assert_code_equal('for i in range(3): a += 1',
                               'for (int i = 0; i < 3; i++) {a += 1;}')

    def test_nested_blocks(self):
        self.assert_code_equal('if a:\n if b: a = b',
                               'if (a) {if (b) {a = b;}}')

    def test_local_var(self):
        self.assert_code_equal('myvar = vec3(0, 0, 0)',
                               'vec3 myvar = vec3(0, 0, 0)')

    def test_local_array(self):
        self.assert_code_equal('arr = Array2[int]',
                               'int arr[2]')

    def test_array_param(self):
        self.assert_code_equal('def fn(arr: Array3[int]): pass',
                               'void fn(int arr[3]) {}')



import ast
from inspect import getsource


def ensure_node_is_function(node):
    """Raise a TypeError if node is not a FunctionDef.

    This isn't strictly necessary but makes errors easier to find.
    """
    if not isinstance(node, ast.FunctionDef):
        raise TypeError('input must be an ast.FunctionDef', node)


def dedent(lines):
    """De-indent based on the first line's indentation."""
    if len(lines) != 0:
        first_lstrip = lines[0].lstrip()
        strip_len = len(lines[0]) - len(first_lstrip)
        for line in lines:
            if len(line[:strip_len].strip()) != 0:
                raise ValueError('less indentation than first line: ' +
                                 line)
            else:
                yield line[strip_len:]


def parse_source(obj):
    src = getsource(obj)
    # TODO(nicholasbishop): this probably doesn't belong here, but
    # it's helpful in the unit tests at least.
    src = '\n'.join(dedent(src.splitlines()))
    return ast.parse(src)


def rename_function(func_node, new_name):
    ensure_node_is_function(func_node)
    func_node.name = new_name


def remove_function_parameters(func_node):
    """Remove all parameters from a function AST node."""
    ensure_node_is_function(func_node)
    func_node.args.args = []


def remove_function_return_type(func_node):
    """Remove return type annotation from a function AST node."""
    ensure_node_is_function(func_node)
    func_node.returns = None



import ast
from shaderdef.ast_util import parse_source

from shaderdef.glsl_var import GlslVar
from shaderdef.glsl_types import vec4

def snake_case(string):
    output = ''
    first = True
    for char in string:
        lower = char.lower()
        if char != lower:
            if first:
                first = False
            else:
                output += '_'
        output += lower
    return output


def _declare_block(block_type, block_name, instance_name, members,
                   array):
    # TODO
    assert len(members) != 0
    assert block_type in ('in', 'out', 'uniform')

    array_string = ''
    if array is not None:
        if isinstance(array, bool):
            array_string = '[]'
        else:
            raise NotImplementedError('only unsized arrays are supported')

    yield '{} {} {{'.format(block_type, block_name)

    for member in members:
        # TODO
        assert not member.name.startswith('gl_')
        yield '    ' + member.declare()

    yield '}} {}{};'.format(instance_name, array_string)


# https://www.opengl.org/wiki/Interface_Block_(GLSL)
class ShaderInterface(object):
    def __init__(self, **kwargs):
        pass

    @classmethod
    def _get_vars(cls):
        # ast is used here instead of inspecting the attributes
        # directly because currently most of the types are just
        # aliases of GlslType rather than subclasses
        src = parse_source(cls)
        cls_node = src.body[0]
        for item in cls_node.body:
            if isinstance(item, ast.Assign):
                name = item.targets[0].id
                # Ignore builtins
                if name.startswith('gl_'):
                    continue
                gtype = item.value.func.id
                interp = None
                if len(item.value.args) == 1:
                    interp = item.value.args[0].id
                yield GlslVar(name, gtype, interpolation=interp)

    @classmethod
    def _declare_block(cls, instance_name, block_type, array=None):
        members = list(cls._get_vars())
        if len(members) == 0:
            return []
        else:
            return list(_declare_block(block_type, cls.block_name(),
                                       instance_name, members, array))

    @classmethod
    def declare_input_block(cls, instance_name, array=None):
        return cls._declare_block(instance_name, 'in', array=array)

    @classmethod
    def declare_output_block(cls, array=None):
        instance_name = snake_case(cls.instance_name())
        return cls._declare_block(instance_name, 'out', array=array)

    @classmethod
    def block_name(cls):
        return cls.__name__

    @classmethod
    def instance_name(cls):
        return snake_case(cls.__name__)


class UniformBlock(ShaderInterface):
    @classmethod
    def declare_input_block(cls, instance_name, array=None):
        return cls._declare_block(instance_name, 'uniform', array=array)


class AttributeBlock(ShaderInterface):
    # For whatever reason GLSL doesn't allow attributes to be
    # aggregated into an interface block
    @classmethod
    def declare_input_block(cls, instance_name=None, array=None):
        if array is not None:
            raise NotImplementedError('attribute arrays not implemented')

        location = 0
        for member in cls._get_vars():
            # TODO(nicholasbishop): correctly handle type size when
            # incrementing location
            yield member.declare_attribute(location)
            location += 1


class FragmentShaderOutputBlock(ShaderInterface):
    # As with attributes, blocks aren't allowed here
    @classmethod
    def declare_output_block(cls, array=None):
        if array is not None:
            raise NotImplementedError('fs output arrays not implemented')

        # TODO(nicholasbishop): dedup
        location = 0
        for member in cls._get_vars():
            # TODO(nicholasbishop): correctly handle type size when
            # incrementing location
            yield member.declare_output(location)
            location += 1


# TODO, builtin
class GlGsIn(ShaderInterface):
    gl_position = vec4()



# Source: https://stackoverflow.com/questions/390250
class EqualityMixin(object):
    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)



import ast

class _LiftAttributes(ast.NodeTransformer):
    def __init__(self, objects):
        self._objects = objects

    # pylint: disable=invalid-name
    def visit_Attribute(self, node):
        value = node.value
        if isinstance(value, ast.Name) and value.id in self._objects:
            return ast.Name(id=node.attr, ctx=node.ctx)
        else:
            return node


def lift_attributes(func_node, objects):
    return _LiftAttributes(objects).visit(func_node)



import ast
from shaderdef.glsl_types import ArraySpec

class Code(object):
    def __init__(self, initial_line=None):
        self.lines = []
        if initial_line is not None:
            self.lines.append(initial_line)
        self._indent_string = '    '

    def __call__(self, line):
        self.lines.append(line)

    def append_block(self, code):
        for line in code.lines:
            line = self._indent_string + line
            if line[-1] not in (';', '{', '}'):
                line += ';'
            self.lines.append(line)

    def one(self):
        if len(self.lines) != 1:
            raise ValueError('expected exactly one line', self)
        return self.lines[0]


def op_symbol(op_node):
    """Get the GLSL symbol for a Python operator."""
    ops = {
        # TODO(nicholasbishop): other unary ops
        ast.UAdd: '+',
        ast.USub: '-',

        # TODO(nicholasbishop): FloorDiv, Pow, LShift, RShift,
        # BitOr, BitXor, BitAnd
        ast.Add: '+',
        ast.Sub: '-',
        ast.Mult: '*',
        ast.Div: '/',
        ast.Mod: '%',

        # Comparison
        ast.Eq: '=',
        ast.NotEq: '!=',
        ast.Lt: '<',
        ast.LtE: '<=',
        ast.Gt: '>',
        ast.GtE: '>=',
    }
    # Python3 matrix multiplication
    if hasattr(ast, 'MatMult'):
        ops[ast.MatMult] = '*'
    return ops[op_node.__class__]


class AstToGlsl(ast.NodeVisitor):
    # pylint: disable=invalid-name,too-many-public-methods
    def visit_Module(self, node):
        if len(node.body) != 1:
            raise NotImplementedError()
        child = node.body[0]
        return self.visit(child)

    def visit_FunctionDef(self, node):
        if getattr(node, 'returns', None) is None:
            return_type = 'void'
        else:
            return_type = self.visit(node.returns).one()
        params = node.args.args[:]
        # Skip self
        if len(params) != 0 and params[0].arg == 'self':
            params = params[1:]
        params = (self.visit(param).one() for param in params)
        code = Code()
        code('{} {}({}) {{'.format(return_type, node.name,
                                   ', '.join(params)))
        for child in node.body:
            code.append_block(self.visit(child))
        code('}')
        return code

    def visit_Pass(self, _):
        # pylint: disable=no-self-use
        return Code()

    def visit_arg(self, node):
        if node.annotation is None:
            raise ValueError('untyped argument: {}'.format(ast.dump(node)))
        adecl = ArraySpec.from_ast_node(node.annotation)
        if adecl is not None:
            return Code('{} {}[{}]'.format(adecl.element_type, node.arg,
                                           adecl.length))

        gtype = self.visit(node.annotation).one()
        return Code('{} {}'.format(gtype, node.arg))

    def visit_Name(self, node):
        # pylint: disable=no-self-use
        return Code(node.id)

    def visit_Attribute(self, node):
        return Code('{}.{}'.format(self.visit(node.value).one(), node.attr))

    @staticmethod
    def is_var_decl(node):
        target = node.targets[0]
        return (isinstance(target, ast.Name) and
                not target.id.startswith('gl_') and
                isinstance(node.value, ast.Call))

    def get_array_decl(self, node):
        # TODO(nicholasbishop): clean mess up
        target = node.targets[0]
        aspec = ArraySpec.from_ast_node(node.value)
        if aspec is None:
            return None
        return Code('{} {}[{}]'.format(aspec.element_type,
                                       self.visit(target).one(),
                                       aspec.length))

    def make_var_decl(self, node):
        target = node.targets[0]
        gtype = node.value.func.id
        return Code('{} {} = {}'.format(gtype,
                                        self.visit(target).one(),
                                        self.visit(node.value).one()))

    def visit_Assign(self, node):
        if len(node.targets) != 1:
            raise ValueError('multiple assignment targets not allowed', node)
        target = node.targets[0]
        if self.is_var_decl(node):
            return self.make_var_decl(node)
        adecl = self.get_array_decl(node)
        if adecl is not None:
            return adecl
        return Code('{} = {}'.format(self.visit(target).one(),
                                     self.visit(node.value).one()))

    def visit_Num(self, node):
        # pylint: disable=no-self-use
        return Code(str(node.n))

    def visit_Call(self, node):
        args = (self.visit(arg).one() for arg in node.args)
        name = self.visit(node.func).one()
        return Code('{}({})'.format(name, ', '.join(args)))

    def visit_Return(self, node):
        return Code('return {}'.format(self.visit(node.value).one()))

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_Str(self, node):
        # pylint: disable=no-self-use,unused-argument
        return Code()

    def visit_UnaryOp(self, node):
        return Code('{}{}'.format(op_symbol(node.op),
                                  self.visit(node.operand).one()))

    def visit_BinOp(self, node):
        return Code('({} {} {})'.format(
            self.visit(node.left).one(),
            op_symbol(node.op),
            self.visit(node.right).one(),
        ))

    def visit_Subscript(self, node):
        return Code('{}{}'.format(self.visit(node.value).one(),
                                  self.visit(node.slice).one()))

    def visit_Index(self, node):
        return Code('[{}]'.format(self.visit(node.value).one()))

    def visit_AugAssign(self, node):
        return Code('{} {}= {}'.format(self.visit(node.target).one(),
                                       op_symbol(node.op),
                                       self.visit(node.value).one()))

    def visit_Compare(self, node):
        if len(node.ops) != 1 or len(node.comparators) != 1:
            raise NotImplementedError('only one op/comparator is supported',
                                      node)
        op = node.ops[0]
        right = node.comparators[0]
        return Code('{} {} {}'.format(self.visit(node.left).one(),
                                      op_symbol(op),
                                      self.visit(right).one()))

    def visit_If(self, node):
        code = Code('if ({}) {{'.format(self.visit(node.test).one()))
        for child in node.body:
            code.append_block(self.visit(child))
        # TODO(nicholasbishop): emit "else if" to make output cleaner
        if len(node.orelse) != 0:
            code('} else {')
            for child in node.orelse:
                code.append_block(self.visit(child))
        code('}')
        return code

    def visit_For(self, node):
        if not isinstance(node.target, ast.Name):
            raise NotImplementedError('for-loop target must be an ast.Name')
        itr = node.iter
        if not isinstance(itr, ast.Call) or itr.func.id != 'range':
            raise NotImplementedError('only range() for loops are supported')
        if len(itr.args) != 1 or not isinstance(itr.args[0], ast.Num):
            raise NotImplementedError('only 0..n for loops are supported')
        end = self.visit(itr.args[0]).one()
        var = self.visit(node.target).one()
        code = Code()
        code('for (int {var} = 0; {var} < {end}; {var}++) {{'.format(var=var,
                                                                     end=end))
        for child in node.body:
            code.append_block(self.visit(child))
        code('}')
        return code

    def visit(self, node):
        ret = super().visit(node)
        if ret is None:
            raise KeyError('unhandled ast node type',
                           ast.dump(node))
        else:
            return ret


def py_to_glsl(root):
    """Translate Python AST into GLSL code.

    root: an ast.FunctionDef object

    Return a list of strings, where each string is a line of GLSL
    code.
    """
    atg = AstToGlsl()
    code = atg.visit(root)
    return code.lines



# pylint: disable=missing-docstring

import ast

class _FindFunctionVisitor(ast.NodeVisitor):
    # pylint: disable=invalid-name
    def __init__(self, function_name):
        super(_FindFunctionVisitor, self).__init__()
        self._function_name = function_name
        self._function = None

    @property
    def function(self):
        return self._function

    def visit_FunctionDef(self, node):
        if node.name == self._function_name:
            self._function = node


def find_function(root, function_name):
    """Search an AST for a function with the given name.

    A KeyError is raised if the function is not found.

    root: an AST node

    function_name: function to search for (string)
    """
    finder = _FindFunctionVisitor(function_name)
    finder.visit(root)

    if finder.function is None:
        raise KeyError('function not found', function_name)
    return finder.function



from shaderdef.stage import Stage


class ShaderDef(object):
    def __init__(self, vert_shader, geom_shader, frag_shader):
        self._vert_shader = Stage(vert_shader)
        self._geom_shader = Stage(geom_shader)
        self._frag_shader = Stage(frag_shader)
        self._library = []

    def add_function(self, func):
        self._library.append(func)

    def translate(self):
        self._vert_shader.translate(self._library)
        self._geom_shader.translate(self._library)
        self._frag_shader.translate(self._library)

    @property
    def vert_shader(self):
        return self._vert_shader.glsl

    @property
    def geom_shader(self):
        return self._geom_shader.glsl

    @property
    def frag_shader(self):
        return self._frag_shader.glsl



#! /usr/bin/env python

from __future__ import print_function

import os
from subprocess import check_call
import sys

from tools.version_util import (SCRIPT_DIR, VERSION_PATH,
                                load_version_as_list,
                                load_version_as_string,
                                format_version_string)


def run_cmd(*cmd):
    """Run a command via subprocess."""
    print(' '.join(cmd))
    check_call(cmd)


def test_and_lint():
    """Run the unit tests and pylint."""
    run_cmd('make', 'test')
    run_cmd('make', 'lint')


def bump_minor_version():
    """Bump the minor version in version.py."""
    version = load_version_as_list()
    print('current version: {}'.format(format_version_string(version)))
    version[-1] += 1
    print('new version: {}'.format(format_version_string(version)))

    contents = "__version__ = '{}'\n".format(format_version_string(version))

    with open(VERSION_PATH, 'w') as wfile:
        wfile.write(contents)


def commit_version():
    run_cmd('git', 'commit', VERSION_PATH,
            '--message', 'Bump minor version number')


def push_branch():
    run_cmd('git', 'push')


def push_tag():
    version = load_version_as_string()
    run_cmd('git', 'tag', version)
    run_cmd('git', 'push', '--tags')


def sdist_and_upload():
    setup_py_path = os.path.join(SCRIPT_DIR, '../setup.py')
    run_cmd(sys.executable, setup_py_path, 'sdist', 'upload')


def main():
    test_and_lint()
    bump_minor_version()
    commit_version()
    push_branch()
    push_tag()
    sdist_and_upload()


if __name__ == '__main__':
    main()



import ast
from typing import (Generic, SupportsAbs, SupportsInt, SupportsFloat,
                    TypeVar)
from shaderdef.equality import EqualityMixin


class GlslType(SupportsAbs, SupportsInt, SupportsFloat):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __getattr__(self, name) -> 'GlslType':
        return GlslType(self, name)

    def __getitem__(self, index) -> 'GlslType':
        return GlslType(self, index)

    def __setitem__(self, index, val) -> 'GlslType':
        return GlslType(self, index, val)

    def __abs__(self) -> 'GlslType':
        return GlslType(self)

    def __add__(self, other) -> 'GlslType':
        return GlslType(self, other)

    def __sub__(self, other) -> 'GlslType':
        return GlslType(self, other)

    def __mul__(self, other) -> 'GlslType':
        return GlslType(self, other)

    def __truediv__(self, other) -> 'GlslType':
        return GlslType(self, other)

    def __int__(self):
        pass

    def __float__(self):
        pass


# pylint: disable=invalid-name
mat2 = GlslType
mat3 = GlslType
mat4 = GlslType
vec2 = GlslType
vec3 = GlslType
vec4 = GlslType

class void(object):
    pass

# TODO
GlslArrayElem = TypeVar('GlslArrayElem')
class GlslArray(Generic[GlslArrayElem]):
    def __init__(self, gtype):
        pass

    def __getitem__(self, index):
        return GlslType(self, index)

    def __setitem__(self, index, val):
        return GlslType(self, index, val)


# TODO(nicholasbishop): this is obviously really ugly, and requires
# users to define their own type alias for larger arrays. But at least
# for now it seems that numbers aren't allowed as type parameters in
# the style that C++ allows (e.g. std::array<float, 3>)
Array1 = GlslArray
Array2 = GlslArray
Array3 = GlslArray
Array4 = GlslArray
Array5 = GlslArray
Array6 = GlslArray
Array7 = GlslArray
Array8 = GlslArray
Array9 = GlslArray
Array10 = GlslArray
Array11 = GlslArray
Array12 = GlslArray
Array13 = GlslArray
Array14 = GlslArray
Array15 = GlslArray
Array16 = GlslArray


class ArraySpec(EqualityMixin):
    """Represents an array declaration.

    This type isn't currently intended to be used by client code
    directly, it's just a convenient form for internal use.
    """
    def __init__(self, element_type, length):
        self.element_type = element_type
        self.length = length

    def __repr__(self):
        return 'ArraySpec({}, {})'.format(self.element_type, self.length)

    @classmethod
    def from_ast_node(cls, node):
        """Create a GlslArray from an AST node if possible.

        If the node cannot be converted then None is returned.
        """
        if not isinstance(node, ast.Subscript):
            return None
        if not isinstance(node.value, ast.Name):
            return None
        name = node.value.id
        prefix = 'Array'
        if not name.startswith(prefix):
            return None
        try:
            num = int(name[len(prefix):])
        except ValueError:
            return None
        if not isinstance(node.slice.value, ast.Name):
            return None
        gtype = node.slice.value.id
        return cls(gtype, num)


# TODO
triangles = 'triangles'
triangle_strip = 'triangle_strip'


# TODO
class noperspective(object):
    pass



from unittest import TestCase

from shaderdef.glsl_types import vec4, noperspective
from shaderdef.interface import ShaderInterface, UniformBlock

class TestShaderInterface(TestCase):
    def test_declare(self):
        class MyBlock(ShaderInterface):
            var = vec4(noperspective)
        self.assertEqual(list(MyBlock.declare_output_block()),
                         ['out MyBlock {',
                          '    noperspective vec4 var;',
                          '} my_block;'])

    def test_uniform(self):
        class MyBlock2(UniformBlock):
            var = vec4()

        self.assertEqual(list(MyBlock2.declare_input_block('x')),
                         ['uniform MyBlock2 {',
                          '    vec4 var;',
                          '} x;'])



import ast
from unittest import TestCase

from shaderdef.glsl_types import ArraySpec


def get_node(string):
    node = ast.parse(string)
    return node.body[0].value


class TestArraySpec(TestCase):
    def test_no_type(self):
        node = get_node('Array4')
        self.assertIs(ArraySpec.from_ast_node(node), None)

    def test_invalid_type(self):
        node = get_node('Array4[1]')
        self.assertIs(ArraySpec.from_ast_node(node), None)

    def test_not_array(self):
        node = get_node('a[3]')
        self.assertIs(ArraySpec.from_ast_node(node), None)

    def test_no_length(self):
        node = get_node('Array[int]')
        self.assertIs(ArraySpec.from_ast_node(node), None)

    def test_valid(self):
        node = get_node('Array4[int]')
        self.assertEqual(ArraySpec.from_ast_node(node),
                         ArraySpec('int', 4))



#! /usr/bin/env python

from setuptools import setup
from tools import version_util


setup(name='shaderdef',
      version=version_util.load_version_as_string(),
      description='Transform Python code into GLSL shaders',
      url='https://github.com/nicholasbishop/shaderdef',
      author='Nicholas Bishop',
      author_email='nicholasbishop@gmail.com',
      license='GNU General Public License v3 or later (GPLv3+)',
      packages=['shaderdef', 'tools'],
      test_suite='test',
      zip_safe=True)



import ast

from shaderdef.rename_ast_nodes import rename_ast_nodes
from test.util import AstTestCase


class TestRenames(AstTestCase):
    def test_rename_assign(self):
        root = ast.parse('a = 3')
        expected = ast.parse('b = 3')
        self.assertEqual(rename_ast_nodes(root, {'a': 'b'}), expected)

    def test_rename_attr_store(self):
        root = ast.parse('foo.a = 3')
        expected = ast.parse('foo.b = 3')
        self.assertEqual(rename_ast_nodes(root, {'a': 'b'}), expected)

    def test_rename_attr_load(self):
        root = ast.parse('foo = bar.a')
        expected = ast.parse('foo = bar.b')
        self.assertEqual(rename_ast_nodes(root, {'a': 'b'}), expected)



import ast

class _Renamer(ast.NodeTransformer):
    # pylint: disable=invalid-name
    def __init__(self, names):
        self.names = names

    def visit_Name(self, node):
        new_name = self.names.get(node.id)
        if new_name is not None:
            node.id = new_name
        return node

    def visit_Attribute(self, node):
        node.value = self.visit(node.value)

        new_name = self.names.get(node.attr)
        if new_name is not None:
            node.attr = new_name

        return node


def rename_ast_nodes(root_node, names):
    return _Renamer(names).visit(root_node)


def rename_gl_builtins(root_node):
    return rename_ast_nodes(root_node, {
        'end_primitive': 'EndPrimitive',
        'gl_position': 'gl_Position',
    })



from unittest import TestCase

from shaderdef.stage import Stage, make_prefix

class TestMakePrefix(TestCase):
    def test_vert_shader(self):
        self.assertEqual(make_prefix('vert_shader'), 'vs_')


class TestAuxFunction(TestCase):
    # pylint: disable=unused-argument
    def test_library(self):
        def simple_stage():
            pass

        def aux(self):
            pass

        stage = Stage(simple_stage)
        code = stage.to_glsl([aux])
        self.assertIn('void aux() {', code)
